/*
      *****************************************************************************
      *  Copyright (C) 2017 by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
*/
// Verilog replacement of schematics module: WRITE_MFM_DECODER_V1CCC
// Clock-recovery is not necessary because a clock, the 4.1MHz clock is
// provided be the RL-Controller ( In contrast to the ST506/ST412 disks ).
// Input of this module is the serial MFM data and the 4.1MHz system clock.
// The write gate signal enables the write circuits and is asserted
// at the start of Preamble PR2, 3 words and only bit 47 = "1".
// 
//             16 15                   0
//  shift --->>+--+--------------------+        LSB first
//  serialin      \--- parallel out ---/ 
//
// 
//  by: Reinhard Heuberger , www.PDP11GY.com, OKT, 2013
// 
//
//   
module my_RL_MFM_DEcoder(
       Enable,                   // Enable
       clk_4_1,                  // RLV-11 Controller clock, 4.1  MHz
       clk_65_6,                 // System Clock 65.6 MHz
       MFM_IN,                   // Serial MFM data from RLV-11 Controller
       WRITE_GATE_I,             // Write Gate In, HiGH if active
                                 // 
       clk_16Bit,                // 16 Bit clock                
       Parallel_out,             // Decoded Data, 16 Bit parallel out
		 Serial_MFM,
		 Serial_strobe,            // Strobe = negative edge , 4.1Mhz clk
       Serial_dat,               // Decoded Data, serial
       WRITE_GATE_O);            // Write-Gate Out 
//
input  Enable, clk_4_1, clk_65_6, MFM_IN, WRITE_GATE_I;
output clk_16Bit, Serial_MFM, Serial_strobe, Serial_dat, WRITE_GATE_O; 
output [15:0] Parallel_out;
//
reg [15:0] latch_16bit;          // Parallel out
reg [16:0] shifter;              // Decoded data shifter
reg [31:0] count_16;             // 16 Bit Counter.
reg [4:0]  synced_WG;            // sync WRITE_GATE to 65.6 MHz clk.
reg [4:0]  sync_MFM;             // sync MFM to 65.6 MHz clk.
reg [4:0]  sync_4_1;             // sync 4.1 MHz rlv-clock to 65.6 MHz clk.
reg synced_clk_16Bit;
reg flag;
reg data_serial;
wire load_L;                     // Set/load signal 16 Bit paralle olut
//
initial
begin
   shifter        <= 0;
   count_16       <= 0;
   flag           <= 0;
end
//
//
//======== Synchronize WRITE_GATE, = WRITE_GATE_I  to FPGA clock ============
//============================================================================
always @ (posedge clk_65_6)
begin
  synced_WG <= { synced_WG[3:0] , WRITE_GATE_I };   // shift
end
assign WRITE_GATE_O = synced_WG[1];
//
//
//============ Synchronize MFM signal, = MFM_IN  to FPGA clock ===============
//============================================================================
always @ (posedge clk_65_6)
begin
  sync_MFM <= { sync_MFM[3:0] , MFM_IN };   // shift
end
assign Serial_MFM = sync_MFM[1];
//
//
//============= Synchronize 4.1 MHz rlv-clock to 65.6 MHz clk ================
//============================================================================
always @ (posedge clk_65_6)
begin
  sync_4_1 <= { sync_4_1[3:0] , clk_4_1 };   // shift
end
assign Serial_strobe = sync_4_1[0];
//
//
//=============== Generate load pulse for parallel out =======================
//============================================================================
always @ (posedge clk_65_6)
begin 
  synced_clk_16Bit = clk_16Bit;
end
assign load_L = ~synced_clk_16Bit & clk_16Bit;
//
//
//-------------------------------  MFM-DECoder ------------------------------
//---------------------------------------------------------------------------
/* 
___--------_______--------_______-----    sync_4_1[0]
_____--------_______--------_______---    sync_4_1[1]
___________--_____________--__________    if(!sync_4_1[0] & sync_4_1[1] )  
___--_____________--____________--____    if(sync_4_1[0] & !sync_4_1[1] )
*/
always @ (posedge clk_65_6)
begin
 if(WRITE_GATE_O & shifter[15] ) begin                // if PR2 @ [15] detected
    flag <= 1;                                        // set flag, = Start
  end
  //
  if( Enable & WRITE_GATE_O ) begin                   // Write Gate set ?
    // 
    //if(( sync_4_1[0] ^ sync_4_1[1] ) & sync_4_1[1]) begin
    if(( clk_4_1 ^ sync_4_1[0] ) & sync_4_1[0]) begin
    //if(!sync_4_1[0] & sync_4_1[1] ) begin           // neg edge  #1
      //data_serial <=  sync_MFM[1];                  // -> Serial out
      data_serial <=  MFM_IN;                         // -> Serial out
    end else if(!sync_4_1[0] & sync_4_1[1] ) begin    // neg edge  #2
      //shifter <= {dat_serial,shifter[16:1]};        //--- shift IN -----
      shifter <= shifter >> 1;                        // shift left to right 
      shifter[16] = data_serial;                      // and set serial in
    end else if(!sync_4_1[1] & sync_4_1[2] ) begin    // neg edge  #3
      if( flag ) begin                                // ***Start****
        count_16 <= count_16 - 1;                     // count
      end else begin
       count_16 <= 0;                                 // clear counter
      end
    end  
   //
  end else begin
    shifter        <= 0;
    count_16       <= 0;
    flag           <= 0;
  end
end
assign clk_16Bit  = count_16[3];
assign Serial_dat = data_serial;
//
//
//=========================== latch 16 bit Output ============================
//============================================================================
always @ (posedge clk_65_6)
begin
  if (!load_L) begin
    latch_16bit[15:0] <= shifter[15:0];           // latch output
  end
end
assign Parallel_out = latch_16bit;               // 16 bit out
//
endmodule
