
/*
      *****************************************************************************
      *  Copyright (C) 2017 by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
*/

//
// Merge of modules: my_send_Drive_Status 
//                   my_receive_Drive_Command
//
// Input of this module is the serial signal Drive_Command, which
// is continuously transmitted from the RL02 controller board. The
// indicator for a valid Drive_Command is the "Marker" Bit 0.
//
// schift >> 15............2...1...0
// serial in                   |   +---> Marker-bit
//                             +-------> 0 = "Seek" mode 
//                                       1 = "Send Drive Status" mode
//
// Output of this module is the 16Bit Drive Status Word. 
//
//  by: Reinhard Heuberger , www.PDP11GY.com, DEZ, 2011
// 
//   NOT blocked version
//
//  V1.0 Implemting Volume Check(VC) handling, Bit 9
//  Improved version for DE10-Lite board,      AUG, 2017
//
// 
module my_command_sequencer_DE10(
       Enable,                   // Enable
       DSbus,                    // Data, 16 Bit parallel in
       Drive_Command,            // Serial incoming Drive Command
       clk_4_1,                  // RLV-11 Controller clock, 4.1  MHz
       clk_65_6,                 // System Clock 65.6 MHz
       seek_ack,                 // Acknowledge head/servo  @ O_ctrl[2]
		 match,
       //
       DCbus,                    // Data, 16 Bit parallel out
       Drive_Status,             // Data, Serial out
       Status_Clock,             // Status-clock Output
       Drive_Ready);             // 
//      
input [15:0] DSbus;    
input  Enable, Drive_Command, clk_4_1, clk_65_6, seek_ack, match;
output [15:0] DCbus;
output Drive_Status, Status_Clock, Drive_Ready;
//
reg [16:0] shifter_in;          // 16+1 bit Drive-Command IN shifter
reg [15:0] shifter_out;         // 16  bit Drive-Status OUT shifter
reg [5:0]  rclksync_4_1;        // like 2 D-FF register,
reg [15:0] DCbus;               // Latches Drive Command word
reg Drive_Ready;                // Drive Ready Low output.
reg detect;                     // Marker detection @ entry 
reg clkctrl;                    // Status Clock control
reg nix;                        //  just used for else-condition
reg VC;                         // Volume Check, Bit 9
wire synced_clk_4_1;            // synchronized status clock
//
initial
begin
   shifter_in     <= 0;
   shifter_out    <= 0;
   Drive_Ready    <= 0;
   DCbus          <= 0;
   detect         <= 0;
   clkctrl        <= 0;
   VC             <= 1;
end
//
//--------------- Synchronize 4.1Mhz clock to FPGA clock --------------------
//---------------------------------------------------------------------------
always @ (posedge clk_65_6)
begin
  rclksync_4_1 <= { rclksync_4_1[2:0] , clk_4_1 };
end
assign synced_clk_4_1 =  rclksync_4_1;
//
//------ Continuously running shiftregister serial_in + parallel out --------
//---------------------------------------------------------------------------
//
always @ (posedge clk_65_6)
begin
  if( Enable & match) begin
   //
   if (Drive_Command) begin                       // Drive_Command @entry 
     detect  <= 1;                                //  set Flag,<= detected
     clkctrl <= 1;                                // Status Clock off
   end  
   if(seek_ack) begin                             // <= Done indicater @ seek 
     Drive_Ready <= 0;                            // clear Drive_Ready
   end
   //+
   if ( rclksync_4_1[1] & ~rclksync_4_1[2] )      // posedge 4.1 MHz
     begin                                        // *start*
      //---------------------------------------------------------------------
      shifter_in  <= {Drive_Command,shifter_in[16:1]};   //--- shift IN -----
      //shifter_out <= {SI,shifter_out[15:1]};           //--- shift OUT ----
      shifter_out <= {Drive_Command,shifter_out[15:1]};  //--- shift OUT ----
      //---------------------------------------------------------------------
      if (shifter_in[0] & detect) begin          // If "Marker" bit0 detection
         detect <= 0;                            // -reset/<=clear detect
         DCbus <= shifter_in[15:0];              // -latch 16bit word
                                                 //
         if (shifter_in[1]) begin                // IF Send Status ?
            if(shifter_in[3]) begin              //   IF Reset-Request ?
              VC <= 0;                           //     clear VC
            end else if(!Enable) begin           //   ELSE if
              VC <= 1;                           //     set VC
            end                                  //
         shifter_out <= DSbus;                   //  than load output SR
         shifter_out[9] <= VC;                   //   set VC
         clkctrl <= 0;                           //  and turn Status clock on. 
         end else begin                          // ELSE  
           //shifter_out <= 16'h0001;            //  clear output SR and send this
                                                 // "Pöpl" bit ( was not ducumented )
           shifter_out <= 16'h0000;
           Drive_Ready <= 1;                     //   -set "seek head/servo" flag
         end
       end else begin
         nix <= nix;
       end
     end else begin
       nix <= nix;
     end
   end else begin
    shifter_in  <= 0;                            // if not enable + match
    shifter_out <= 0;                            // if not enable + match
    VC          <= 1;
   end
end
//
assign Drive_Status = shifter_out[0];
assign Status_Clock =(~clkctrl ) & (~synced_clk_4_1);
//
endmodule
