
/*
      *****************************************************************************
      *  Copyright (C) 2017 by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
*/  

//
// Ersetzt die beiden asychronen Module
//   "my_little_idexer"   und   "my_little_address_maker"
//
// Dieses Modul generiert die Sector Addressen und die Addressen innerhalb 
// eines Sectors für einen Track.
// Ein Track besteht aus 40 Sectoren(0-39) und ein Sector besteht jeweils
// aus 140 16-Bit Wörtern inclusive servo/header/CRC Information. Im FPGA
// wird ein Sector mit 144 16-Bit wörtern zwecks 16Bit-boundary dargestellt.
//
// "little" Version, da Cyclone II EB2C20F484C7 nur 52 4K blöcke hat und
// somit nur ein Track mit 5760 16-bit words im FPGA gehalten werden kann.
// timing:
// sector  :  -------________---------------------------------------_____
// ENC_clk :  |||||          |||||||||||||||||||||||||||||||||||||
// Data_end:  -----__---------------------------------------------__-----
// SEC_clk_L:--------|----------------------------------------------|-----
//                           <----- Data, 140 16-Bit words ----->
//
//
//  by: Reinhard Heuberger , www.PDP11GY.com , JAN, 2012
//  improved version for DE10-Lite board,      AUG, 2017
//
module my_little_sequencer_DE10 ( 
       enable,		          // Enable Input
       ENC_clk,		          // MFM-ENcoder 16Bit-clock
       sector, 		          // Sector pulse-clock
       clk_65_6,              // System Clock 65.6 MHz
                              //
       address_out,           // Calculated address of data
       data_end,              // Data End indicator
       sector_0,              // sector 0 indicator
       tmp);                  // for measurement purpose only.
//   
input enable, ENC_clk, sector, clk_65_6;
output [12:0] address_out;
output data_end, sector_0, tmp;
//
reg [31:0] current_address;   // augenblickliche Addresse^
reg [31:0] grenze_unten;      // Start Addresse ( DP RAM )
reg [31:0] grenze_oben;       //  bis g�ltige Obergrenze.
reg [31:0] sector_counter;    // Sector counter
reg [12:0] address_out;       // 13-Bit output Adress to Dual-ported Ram
reg [3:0]  synced_SEC_clk;    // like 2 D-FF register, sync to 50MHz clk.
reg [3:0]  synced_ENC_clk;    // like 2 D-FF register, sync to 50MHz clk.
reg data_end;                 // date tranmition end_L in one sector
reg sector_0;                 // sector #0 detection
//
initial
begin
   current_address  = 0;
   grenze_unten     = 0;
   grenze_oben      = 140;
end
//
//
//======== Synchronize 16Bit clock, = ENC_clk clock to FPGA clock ============
//============================================================================
always @ (posedge clk_65_6)
begin
  synced_ENC_clk <= { synced_ENC_clk[2:0] , ENC_clk };   // shift
end
//
//
//======== Synchronize sector clock, = SEC_clk to FPGA clock =================
//============================================================================
always @ (posedge clk_65_6)
begin
  synced_SEC_clk <= { synced_SEC_clk[2:0] , sector };   // shift
end
//
//
//================ Sector Idex-Adressen bilden, 0-39 ========================
//========= 16Bit Adressen innerhalb eines Sector bilden, 0-139 =============
//===========================================================================
always @ (posedge clk_65_6)
begin
   if ( ~synced_SEC_clk[1] & synced_SEC_clk[2] )      // @ negedge SEC clk
   //if ( synced_SEC_clk[1] & ~synced_SEC_clk[2] )      // @ posedge SEC clk
    begin
     data_end <= 1;
     if(enable) begin
      sector_counter = sector_counter +1;          // Increment
       if ( sector_counter == 40 ) begin           //  If Limit ( 0-39 )    
          sector_0 = 1;   
          sector_counter  = 0;                     //      clear counter    
          current_address = 0;                     //      and address
          grenze_unten    = 0;
          grenze_oben = 140;                       //  set upper limit 
       end else if( sector_counter < 40 ) begin
          sector_0 = 0;
          grenze_unten = grenze_unten + 144;       //  DP-RAM-Index = +144
          current_address = grenze_unten;          //  Neue untere Grenze
          grenze_oben = grenze_unten + 140;        //  set upper limit  
       end    
     end   
   end
   //
   // Synchron zur 50Mhz clock werden nun mit der negativen Flanke der
   // 16Bit clock die Adressen innerhalb eines Sectors gebildet.
   //      
   if ( ~synced_ENC_clk[1] & synced_ENC_clk[2] )      // @ negedge ENC clk
   //if ( synced_ENC_clk[1] & ~synced_ENC_clk[2] )    // @ posedge ENC clk
    begin                                             // *start*
     if(enable) begin
       if ( current_address == grenze_oben+2) begin   // $....
         data_end <= 0;                               // All done=data_end_L
        end else begin
         current_address = current_address + 1;       // words/sector-counter
         address_out = current_address[12:0];         // send Adress 
         data_end <= 1;              
       end
     end else begin
       current_address = 0;
       grenze_unten    = 0;
       grenze_oben   = 140;
       data_end       <= 0; 
     end
   end    
end
//
endmodule
