/*
              *****************************************************************************
              *  Copyright (C) 2017 by Reinhard Heuberger                                 *
              *                                                                           *
              *  All rights reserved.                                                     *
              *                                                                           *
              *  Redistribution and use in source and binary forms, with or without       * 
              *  modification, are permitted provided that the following conditions       * 
              *  are met:                                                                 *
              *                                                                           * 
              *  1. Redistributions of source code must retain the above copyright        * 
              *     notice, this list of conditions and the following disclaimer.         *
              *  2. Redistributions in binary form must reproduce the above copyright     *
              *     notice, this list of conditions and the following disclaimer in the   *
              *     documentation and/or other materials provided with the distribution.  *
              *  3. Neither the name of the author nor the names of its contributors may  * 
              *     be used to endorse or promote products derived from this software     *
              *     without specific prior written permission.                            *
              *                                                                           *
              *****************************************************************************
*/
// This module generates the RL02 sector pulse. We need 2 different
// sector pulse. Pulse #1 is responsible for the Data start phase if
// this pulse is LOW. Pulse #2 is responsible for the Servo/Header
// start time if this pulse is LOW. Here is the required timing:
//
//      |<---Sector-time,625uSec = 16KHz-->|
// _____-----------_________________________-----------_______
//      |
//     >|---------|<=== Sector-Puls #1 = 62.5 uSec
//     >|------|<====== Pulse 48.5 uSec
//     >|-------|<===== Pulse 57.4 uSec
//
// 
//  by: Reinhard Heuberger , www.PDP11GY.com , MAR 2013
//      Improved Version for DE10-Lite board,  AUG 2017
//
//
module my_sector_pulse_DE10 ( 
	                          clk_65_6,
	                          Pulse_62_5uSec_H,
	                          Pulse_48_5uSec_H,
	                          Pulse_48_5uSec_L,
	                          Pulse_57_4uSec_H,
	                          Pulse_57_4uSec_L);
//
input  clk_65_6;
output Pulse_62_5uSec_H;
output Pulse_48_5uSec_H, Pulse_48_5uSec_L; 
output Pulse_57_4uSec_H, Pulse_57_4uSec_L;
//
reg baseline, flip;
reg Pulse_62_5uSec_H, Pulse_62_5uSec_L; 
reg Pulse_48_5uSec_H, Pulse_48_5uSec_L;
reg Pulse_57_4uSec_H, Pulse_57_4uSec_L;
reg [31:0] count_base;
reg [31:0] count_48_5uSec;
reg [31:0] count_57_4uSec;
//
// 
// internal : Construct the baseline sector clock, 1:1 duty-cyle. 
// 65.6 MHz = 0.01524 uSec , 625 uSec = 2 mal 312.5 uSEC 
// Also 312.5 uSec / 0.01524 = 20505.24934 
// Counter läuft aber von 0, also counter-wert 20506
//
always @ ( posedge clk_65_6 ) begin
    if ( count_base <= 20506 ) begin
       count_base = count_base + 1;
    end else begin
       count_base = 0;
       flip = ~flip;
    end
end
//
always @ ( posedge clk_65_6 ) 
begin
    Pulse_62_5uSec_H  <= ~flip;
end
// 
//     Construct 2 additional sector pulse based on :
// 16bit period = 3.917uSec , sector periode = 0.625mSec
//
// a)65.6 MHz = 0.01524 uSec , also:
//   (48.5uSec : 0.01524) = 3183 Counter-wert.
always @ ( posedge clk_65_6 ) begin
  if (Pulse_62_5uSec_H) begin
    if ( count_48_5uSec <= 3183 ) begin
       count_48_5uSec = count_48_5uSec + 1;
       Pulse_48_5uSec_H <= 1;
       Pulse_48_5uSec_L <= 0;
    end else begin
       count_48_5uSec = count_48_5uSec;
       Pulse_48_5uSec_H <= 0;
       Pulse_48_5uSec_L <= 1;
    end
  end else if (!Pulse_62_5uSec_H) begin
      count_48_5uSec <= 0;
      Pulse_48_5uSec_H <= 0;
      Pulse_48_5uSec_L <= 1;
  end
end
//
// b) 65.6 MHz = 0.01524 uSec , also:
//  (57_4uSec : 0.01524) = 3768 Counter-wert.
always @ ( posedge clk_65_6 ) begin
  if (Pulse_62_5uSec_H) begin
    if ( count_57_4uSec <= 3768 ) begin
       count_57_4uSec = count_57_4uSec + 1;
       Pulse_57_4uSec_H <= 1;
       Pulse_57_4uSec_L <= 0;
    end else begin
       count_57_4uSec = count_57_4uSec;
       Pulse_57_4uSec_H <= 0;
       Pulse_57_4uSec_L <= 1;
    end
  end else if (!Pulse_62_5uSec_H) begin
      count_57_4uSec <= 0;
      Pulse_57_4uSec_H <= 0;
      Pulse_57_4uSec_L <= 1;
  end
end 
endmodule